import { fromBase64Url, toBase64Url } from "./base64";

// === 常量区域 ===
const QUERY_YGO_TYPE = 'ygotype';
const QUERY_VERSION = 'v';
const ARG_DECK = 'deck';
const QUERY_DECK = 'd';
const QUERY_NAME = 'name';

const URL_SCHEME_HTTP = 'http';
const URL_HOST_DECK = 'deck.ourygo.top';

// === 工具函数 ===
function toBinary(value: number, length: number): string {
  return value.toString(2).padStart(length, '0');
}

function encodeCards(cards: number[]): string {
  const bits: string[] = [];
  for (let i = 0; i < cards.length; ) {
    const id = cards[i];
    let count = 1;
    while (i + count < cards.length && cards[i + count] === id && count < 3) count++;
    const prefix = count === 2 ? '10' : count === 3 ? '11' : '01';
    bits.push(prefix + toBinary(id, 27));
    i += count;
  }
  return bits.join('');
}

function countUnique(cards: number[]): number {
  let num = 0;
  for (let i = 0; i < cards.length; i++) {
    const id = cards[i];
    if (id > 0) {
      num++;
      if (i < cards.length - 1 && cards[i + 1] === id) {
        i++;
        if (i < cards.length - 1 && cards[i + 1] === id) {
          i++;
        }
      }
    }
  }
  return num;
}

// === 主函数 ===

export function toYGOMobileDeckURL(
  main: number[],
  extra: number[],
  side: number[],
  customParams: Record<string, string> = {},
): string {
  const mNum = countUnique(main);
  const eNum = countUnique(extra);
  const sNum = countUnique(side);

  const header = toBinary(mNum, 8) + toBinary(eNum, 4) + toBinary(sNum, 4);
  let bitString = header + encodeCards(main) + encodeCards(extra) + encodeCards(side);
  while (bitString.length % 8 !== 0) bitString += '0';

  const bytes = new Uint8Array(bitString.length / 8);
  for (let i = 0; i < bytes.length; i++) {
    bytes[i] = parseInt(bitString.slice(i * 8, i * 8 + 8), 2);
  }

  let encoded = toBase64Url(bytes)

  const searchParams = new URLSearchParams();
  for (const [key, value] of Object.entries(customParams)) {
    searchParams.set(key, value);
  }
  searchParams.set(QUERY_YGO_TYPE, ARG_DECK);
  searchParams.set(QUERY_VERSION, '1');
  searchParams.set(QUERY_DECK, encoded);

  return `${URL_SCHEME_HTTP}://${URL_HOST_DECK}?${searchParams.toString()}`;
}


export function fromYGOMobileDeckURL(uri: string): {
  main: number[];
  extra: number[];
  side: number[];
  name?: string;
} {
  const url = new URL(uri);
  if (url.searchParams.get(QUERY_YGO_TYPE) !== ARG_DECK) {
    throw new Error('Not a YGO Mobile deck URI');
  }

  let encoded = url.searchParams.get(QUERY_DECK);
  if (!encoded) throw new Error('Missing deck data');

  const bytes = fromBase64Url(encoded);

  const bits = Array.from(bytes)
    .map((b) => b.toString(2).padStart(8, '0'))
    .join('');

  const mNum = parseInt(bits.slice(0, 8), 2);
  const eNum = parseInt(bits.slice(8, 12), 2);
  const sNum = parseInt(bits.slice(12, 16), 2);

  const all = mNum + eNum + sNum;
  const cards: number[] = [];
  let pos = 16;
  
  const res = {
    main: [] as number[],
    extra: [] as number[],
    side: [] as number[],
  }

  let i = 0;

  while (pos + 29 <= bits.length && cards.length < all) {
    const countBits = bits.slice(pos, pos + 2);
    const count = countBits === '10' ? 2 : countBits === '11' ? 3 : 1;
    const id = parseInt(bits.slice(pos + 2, pos + 29), 2);
    const field = i < mNum ? 'main' : i < mNum + eNum ? 'extra' : 'side';
    res[field].push(...Array(count).fill(id));
    pos += 29;
    i++;
  }

  return {
    ...res,
    name: url.searchParams.get(QUERY_NAME),
  };
}
